    program fewmarun
!
!   Program for general MEWMA run length calculations
!   Calculates the in- and out-of-control ARL for a FEWMA
!   If requested, estimates the control limit for a target ARL
!

    implicit none
    real(8), allocatable :: sigma(:,:), gmsig(:,:), rmat(:,:), rbar(:,:), &
                            gminv(:,:), facsig(:,:), facstd(:,:), rsigr(:,:), &
                            data(:), delta(:), rdelt(:), stedsl(:,:), &
                            dew(:), dewdl(:), work(:), work2(:), work3(:), &
                            work4(:)

    real(8) ::  r, c, den, od, tsq, t2sq, tsqmax, x, y, xbar, ybar, &
                xbaro, ybaro, &
                ssx, ssy, sxy, ssxo, ssyo, sxyo, devx, devy, fn, fn1, corrf, &
                cut, slope, tararl, baseh, predh, se, fno, fno1,  &
                resvar, comter, alow, ahi, cuto, slopeo, resvro, predo, seo, &
                sumdel, flam, flamz, dflam, run, runic, runooc, &
                cter, ater, bter, disc, &
                zero=0.d0, one=1.d0, two=2.d0, four=4.d0, half=0.5d0, &
                numfac=5.d0, denfac= 25.d0, einv=0.368d0, oneein=0.632d0

    integer :: nord, npar, i, j, k, m, ij, kl, inx=100000, nravl, mravl, ii, &
               nsim, isim, ngap, stuck=50000

    character(60) :: filename
    logical :: steady, whchmt, over, gotin, gotout, echo, design, tripper, report
    character(1) :: answer

    save inx

    write(*,*) 'Program to find the control limit for the '
    write(*,*) 'general multivariate EWMA and/or its performance'
    write(*,1001) 'Give the name of the file to write results'
1001 format(1x,a/)
    read(*,'(a)') filename
    open(11,file=filename)
    write(11,*) ' Program to find the control limit for the '
    write(11,*) 'general multivariate EWMA and/or its performance'
    write(*,1001) 'Give the number of dimensions   :  '
    read(*,*) nord
    write(11,1001) 'Give the number of dimensions'
    write(11,*) nord
    npar = nord * (nord+1) / 2

    allocate (sigma(nord,nord), gmsig(nord,nord), rmat(nord,nord), &
             rbar(nord,nord), gminv(nord,nord), facsig(nord,nord), &
             facstd(nord,nord), rsigr(nord,nord), data(nord), delta(nord), &
             rdelt(nord), dew(nord), dewdl(nord), &
             stedsl(npar, npar+2), work(npar), &
             work2(npar), work3(npar), work4(npar))



    write(*,1001) 'What r and c values do you want to use?  :  '
    read(*,*) r, c
    write(*,1001) '(you only need to give the lower triangle)'
    write(11,1001) 'What r and c values do you want to use?'
    write(*,'(2f8.3)') r, c
    write(11,'(2f8.3)') r, c

    den = one + (nord - one) * c
    od = r * c / den

    write(*,*) 'Now give the elements of the covariance matrix'
    write(11,*) 'Now give the elements of the covariance matrix'
    write(*,1001) '(you only need to give the lower triangle)'
    write(11,*) '(you only need to give the lower triangle)'
    do i = 1, nord
      write(*,'(/a,i4,a)') ' Row',i,' :  '
      read(*,*) (sigma(i,j),j=1,i)
      write(11,'(10f8.3)') (sigma(i,j),j=1,i)
      rmat(i,i) = r / den
      rbar(i,i) = one - r / den
      do j = 1, i-1
        sigma(j,i) = sigma(i,j)
        rmat(i,j) = od
        rmat(j,i) = od
        rbar(i,j) = -od
        rbar(j,i) = -od
        end do
      end do

!
!   Get R sigma R'
!

    rsigr = matmul(rmat, matmul(sigma, rmat))

!
!   Get triangular factor of sigma
!
    call matfac(sigma,facsig,nord)

    write(* ,1001) 'Do you want initial state or steady state?  :  '
    write(11,*) 'Do you want initial state or steady state?'
    read(*,'(a)') answer
    steady = answer .eq. 'S' .or. answer .eq. 's'
    if (steady) then
      write(11,*) ' Steady state'
      write( *,*) ' Steady state'
      else
      write( *,*) ' Initial state'
      write(11,*) ' Initial state'
      endif

!
!     Get asymptotic covariance matrix
!
    nravl = 0
    do i = 1, nord
      do j = i, nord
        nravl = nravl + 1
        do k = 1, npar
          stedsl(nravl,k) = zero
          end do
        stedsl(nravl,nravl) = one
        work3(nravl) = rsigr(i,j)
        work4(nravl) = zero
        mravl = 0
        do k = 1, nord
          mravl = mravl + 1
          stedsl(nravl,mravl) = stedsl(nravl,mravl)-rbar(i,k)*rbar(j,k)
          do m = k+1, nord
            mravl = mravl + 1
            stedsl(nravl,mravl) = stedsl(nravl,mravl) - rbar(i,k) * &
              rbar(j,m) - rbar(i,m) * rbar(j,k)
            end do
          end do
        end do
      end do

    echo = .false.
    call solve(stedsl,work3,work4,work,work2,npar,echo)
    nravl = 0
    write( *,'(/'' Steady state covariance matrix'')')
    write(11,'(/'' Steady state covariance matrix'')')
    do i = 1, nord
      do j = i, nord
        nravl = nravl + 1
        gmsig(i,j) = work(nravl)
        gmsig(j,i) = work(nravl)
        end do
      write( *,'(8f9.4)') gmsig(i,:)
      write(11,'(8f9.4)') gmsig(i,:)
      end do

    call matfac(gmsig,facstd,nord)

    write(*,*) 'Do you want to find h or do you already know it?'
    write(*,1001) '(answer Yes to get  h or No to do the  OOC ARL)'
    read(*,'(a)') answer
    design = answer .eq. 'Y' .or. answer .eq. 'y'
    write(11,*) ' Do you want to find h or do you already know it?'
    write(11,*) ' (answer Yes to get  h or No to do the  OOC ARL)'
    if (design) then
      write(11,*) ' Yes - get h and OOC ARL'
      baseh = nord
      write(*,1001) ' What is the null ARL?'
      read(*,*) tararl
      write(11,1001) ' What is the null ARL?'
      write(11,'(f10.0)') tararl
      stuck = 50 * tararl
      else
      write(11,*) ' No - I have h, just get the OOC ARL'
      write(*,1001) ' What is the in-control h?'
      write(11,*) ' What is the in-control h?'
      read(*,*) baseh
      write(11,'(f10.3)') baseh
      endif
    write(*,1001) 'Now give the vector of shifts in the mean'
    write(11,1001) 'Now give the vector of shifts in the mean'
    read(*,*) (delta(i),i=1,nord)
    write(11,'(10f8.4)') (delta(i),i=1,nord)
!
!     Get noncentrality
!
    call matvec(rmat, delta, rdelt, nord)

    sumdel = sum(delta)

    do i = 1, nord
      work(i) = (delta(i) - dot_product(facsig(i,:i-1), work (:i-1)))/facsig(i,i)
      work2(i)= (delta(i) - dot_product(facstd(i,:i-1), work2(:i-1)))/facstd(i,i)
      end do
    flam = sqrt(sum(work(:nord)**2))
    write(* ,'('' This gives root noncentrality   '',f9.3)') flam
    write(11,'('' This gives root noncentrality   '',f9.3)') flam
    flamz = sum(work2(:nord)**2)
    dflam = flam * sqrt((two - r) / r)
    write(* ,'('' DEWMA steady state noncentrality'',f9.3)') dflam
    write(11,'('' DEWMA steady state noncentrality'',f9.3)') dflam
    flamz = sqrt(flamz)
    write(11,'('' FEWMA steady state noncentrality'',f9.3)') flamz
    write(* ,'('' FEWMA steady state noncentrality'',f9.3)') flamz

    write(*,1001) 'How many runs do you want simulated?'
    write(11,1001) 'How many runs do you want simulated?'
    read(*,*) nsim
    write(11,*) nsim
    write(*,1001) 'Give two integer seeds less than 30000'
    read(*,*) ij,kl
    write(11,1001) 'Give two integer seeds less than 30000'
    write(11,'(2i7)') ij,kl
    ngap = nsim / 20
    write(*,*) '.....Run starting.....'

    tripper = .not. steady

    xbar = zero
    ybar = zero
    ssx  = zero
    ssy  = zero
    sxy  = zero
    fn   = zero
    xbaro = zero
    ybaro = zero
    ssxo  = zero
    ssyo  = zero
    sxyo  = zero
    fno   = zero
    over = .false.
    echo = .false.
arun:   do  isim = 1, nsim
      dew   = zero
      dewdl = zero

      if (.not. steady) gmsig = zero

      run = zero
      tsqmax = zero
      gotin  = .not. design
      gotout = .false.
!
!     Initialize
!
      predh = baseh
      dew = zero
      if (steady) then
        tripper = .false.
        do inx = 1, 100     ! Up to 100 tries to get a feasible initial steady-state start
          call randn(data, nord, ij, kl)
          if (inx > 90) write(*,'('' try'',i5, '' h and norm'',2f8.2, &
            & '' data'',10f5.2)') inx, baseh, sum(data**2), data
          if (sum(data**2) < baseh) then
            tripper = .true.
            call matvec(facstd, data, dew, nord)
            exit
            end if
          end do
        if (.not. tripper) then
          write(*,*) '100 tries have failed to get a feasible start.  Exiting'
          stop
          end if
        end if
!
!     Loop point for generating new case
!
acase: do

        run = run + one
        if (run .gt. stuck) then
          write(*,*) 'Seems to be stuck. sim', isim, ' run',run,' h',baseh,&
            ' tsq', tsq,' tsqmax',tsqmax,' t2sq',t2sq
          write(*,1002) 'gmsig',gmsig
          write(*,1002) 'gminv',gminv
          write(*,1002) 'data',(data(ii),ii=1,nord)
          write(*,1002) 'dew',(dew(ii),ii=1,nord)
          write(*,1002) 'delta',(delta(ii),ii=1,nord)
          write(*,1002) 'work ',work(:nord)
          write(*,1002) 'work2',work2(:nord)
 1002     format(/1x,a,(10g14.7))
          stop
          endif
        call randn(data, nord, ij, kl)        ! Get random N(0,I)
        call matvec(facsig, data, data, nord) ! Transform to N(0,Sigma)
        call matvec(rbar, dew, work, nord)    ! First term of FEWMA
        call matvec(rmat, data, work2, nord)  ! Second term of FEWMA
        dew = work(:nord) + work2(:nord)      ! New FEWMA

        call matvec(rbar, dewdl, work, nord)  ! and the updated OOC shift
        dewdl = work(:nord) + rdelt(:nord)

        if (.not. steady) gmsig = matmul(rbar, matmul(gmsig, rbar)) + rsigr

        gminv = gmsig

        call solve(gminv, dew, dewdl, work, work2, nord, echo)
        tsq  = dot_product(work(:nord), dew)
        t2sq = dot_product(dew - dewdl, work(:nord) - work2(:nord))
        if (run == 50000) then    ! Report and bail out
          write(*,'('' at 50k.  h'', f8.3,'' tsq'', 2f9.3)') baseh, tsq, t2sq
          cycle
          end if
        tsqmax = max(tsq,tsqmax)

        if (.not. gotin) then
          if (tsq .gt. baseh) then
            gotin = .true.
            runic = run
            y =    (run)
            x = baseh
            corrf = numfac / (float(isim) + denfac) + 0.02
!
!         Do Robbins-Monro-style update
!
          if (run .lt. tararl) then
            baseh = baseh * (one + einv * corrf)
            else
            baseh = baseh * (one - oneein * corrf)
            over = .true.
            endif
!
!         If we are beyond the warmup, then update the statistics
!
          if (isim > 50) then
            devx = x - xbar
            devy = y - ybar
            fn1 = fn + one
            xbar = xbar + devx / fn1
            ybar = ybar + devy / fn1
            ssx = ssx + devx * (x - xbar)
            ssy = ssy + devy * (y - ybar)
            sxy = sxy + devx * (y - ybar)
            fn = fn1
            endif
!
!         If we are beyond try 100, use regression to update baseh
!
          if (isim > 100 .and. (min(mod(isim,ngap),mod(isim, 4))) == 0) then
            slope = sxy / ssx
            cut = ybar - slope * xbar
            predh = (tararl - cut) / slope
            resvar = (ssy - sxy**2 / ssx) / (fn - two)
            comter = four * resvar / ssx
            ater = slope ** 2 - comter
            bter = two * (slope * (cut - tararl) + comter * xbar)
            cter = (cut - tararl) ** 2 - four * resvar * (one / fn + &
              xbar ** 2 / ssx)
            disc = bter ** 2 - four * ater * cter
            if (disc .lt. zero) baseh = predh
            end if
          endif
        end if

      if (.not. gotout) then                   ! Update OOC statistics
        if (t2sq .gt. baseh) then
          runooc = run
          gotout = .true.
          if (isim > 50) then     ! baseh varying
            x = baseh
            y = run
            runooc = run
            devx = x - xbaro
            devy = y - ybaro
            fno1 = fno + one
            xbaro = xbaro + devx / fno1
            ybaro = ybaro + devy / fno1
            ssxo = ssxo + devx * (x - xbaro)
            ssyo = ssyo + devy * (y - ybaro)
            sxyo = sxyo + devx * (y - ybaro)
            fno = fno1
            endif
          endif
        end if
        if (gotin .and. gotout) exit acase
        end do acase

      if (mod(isim, ngap) == 0) then
        write(* ,'(/'' runs completed'',i8)') isim
        write(11,'(/'' runs completed'',i8)') isim
        if (design) then

          if (disc .lt. zero) then
             write(*,101) predh
             write(11,101) predh
             else
             disc = sqrt(disc)
             alow = (-bter - disc) / (two * ater)
             ahi  = (-bter + disc) / (two * ater)
             write(*,101) predh,alow,ahi
             write(11,101) predh,alow,ahi
             baseh = predh
             endif
 101      format(' Estimated h is',f9.3,:,' with CI',2f9.3)

          slopeo = sxyo / ssxo
          cuto = ybaro - slopeo * xbaro
          resvro = (ssyo - sxyo**2 / ssxo) / (fno - two)
          predo = cuto + slopeo * predh
          seo = sqrt(resvro * (one / fno + (predh - xbaro) ** 2 / ssxo))
          write(*,102) predo,predo-two*seo,predo+two*seo
          write(11,102) predo,predo-two*seo,predo+two*seo
 102      format(' OOC ARL is',f9.3,'  with CI',2f9.3)

          else
          se  = sqrt(ssy  / (fn  * (fn  - one)))

!         By changing initialization of 'gotin', you can calculate
!         the in-control ARL also and print it as below.
!
!         write(* ,104) 'IC ', ybar , ybar -two*se , ybar + two*se
!         write(11,104) 'IC ', ybar , ybar -two*se , ybar + two*se
 104      format(1x,a,' ARL is',f9.3,'  with CI',2f9.3)
          seo = sqrt(ssyo / (fno * (fno - one)))
          write(* ,104) 'OOC', ybaro, ybaro-two*seo, ybaro+two*seo
          write(11,104) 'OOC', ybaro, ybaro-two*seo, ybaro+two*seo
          endif

        end if

      end do arun
    end program fewmarun


!==========================================================================
   subroutine solve(ain, y1, y2, x1, x2, nord, echo)
!
!   Subroutine to solve a system of linear equations for two right sides
!
    implicit none
    real(8) ::  ain(nord,nord), a(nord,nord+2), x1(nord), x2(nord), &
                y1(nord), y2(nord), pivot
    integer :: i, j, nord
    logical echo

    a(:,:nord) = ain
    a(:,nord+1) = y1
    a(:,nord+2) = y2
    if (echo) then
    do i = 1, nord
      write(*,'('' B4'',(8f10.4))') (a(i,j),j=1,nord+2)
      end do
    end if

    do i = 1, nord-1
      do j = i+1, nord
        pivot = a(j,i) / a(i,i)
        a(j,i:nord+2) = a(j, i:nord+2) - a(i, i:nord+2) * pivot
        end do
      end do

    if (echo)  write(*,*) 'Solution'
    do i = nord,1,-1
      x1(i) = (a(i,nord+1) - dot_product(a(i,i+1:nord), x1(i+1:nord))) / a(i,i)
      x2(i) = (a(i,nord+2) - dot_product(a(i,i+1:nord), x2(i+1:nord))) / a(i,i)
      end do

    if (echo) then
      write(*,'(1x,7f11.5)') x1
      write(*,*)
      write(*,'(1x,7f11.5)') x2
      endif
    echo = .false.
    return
    end

!==========================================================================
    subroutine matfac(sigma,factor,nord)
!
!   Subroutine to find the lower triangular factor of a  symmetric matrix
!
      implicit none
      real(8) :: sigma(nord,nord), factor(nord,nord), sumel, rutarg, &
                 zero=0.d0
      integer :: i, j, k, nord

      factor(1,1) = sqrt(sigma(1,1))

      do i = 2, nord
        do j = 1, i
          sumel = dot_product(factor(i,:j-1), factor(j,:j-1))
        if (j .lt. i) then
          factor(i,j) = (sigma(i,j) - sumel) / factor(j,j)
          factor(j,i) = zero
          else
          rutarg = sigma(i,i) - sumel
          if (rutarg .le. zero) then
            write(11,'('' Error in MATFAC.  Variable'',i5,'' rutarg'', &
              & g15.7)') i,rutarg
            write(11,*) ' Matrix'
            write(*,'('' Error in MATFAC.  Variable'',i5,'' rutarg'',  &
              & g15.7)') i,rutarg
            write(*,*) ' Matrix'
            do  k = 1, nord
              write( *, '(10f8.3)') sigma(k,:)
              write(11, '(10f8.3)') sigma(k,:)
              end do
            write(*,*) ' Factor so far'
            write(11,*) ' Factor so far'
            do k = 1, i
              write( *, '(10f8.3)') factor(k,:k)
              write(11, '(10f8.3)') factor(k,:k)
              end do
            rutarg = abs(rutarg)
            endif
          factor(i,i) = sqrt(rutarg)
          endif
        end do
      end do
    return
    end subroutine matfac


!===========================================================================
    subroutine randn(x,nord,ij,kl)
!
!   Procedure for simulating vectors of N(0,1) variables
!   Algorithm based on Ripley International Statistical Review
!   Calls ranmar to get random uniforms
!
    implicit none
    integer, parameter :: nuni = 200
    real(8) :: x(nord), work(nuni), u, v, z, &
               quart=0.25d0, half=0.5d0, one=1.d0, &
               zero=0.d0, cons=1.71552777d0, c=0.259d0, onep=-0.350927217d0
    integer :: index, i, j, ij, kl, nord, inrej
    save work,index

    index = nuni

    do i = 1, nord
rejec: do inrej = 1, 200
        index = index + 2
        if (index .gt. nuni-1) then
          call ranmar(work, nuni, ij, kl)
          index = 1
          endif
        u = work(index)
        if (u .eq. zero) then               ! Kludge needed to avoid zero divide
          u = 1.d-10
          endif
        v = cons * (work(index+1) - half)
        x(i) = v / u
        z = quart * x(i) ** 2
        if (inrej > 190) then
          write(*,'('' reject'',i3,'' work'',2f8.5,'' u, v, x, z, tests'',6f7.3)') &
            inrej, work(index:index+1), u, v, x(i), z, c/u-onep, -log(u)
          end if
        if (z .gt. one - u) then
          if (z .gt. c / u - onep) cycle rejec
          if (z .gt. -log(u)) cycle rejec
          endif
        exit
        end do rejec
      end do
    return
    end

!===============================================================================
    subroutine ranmar(rvec, len, ij, kl)
!
!     this is the random number generator proposed by george marsaglia in
!     florida state university report: fsu-scri-87-50
!     it was slightly modified by f. james to produce an array of pseudorandom
!     numbers.
!
    implicit none
    real(8) :: rvec(len), u(97), c, cd, cm, s, t, uni, &
               zero=0.d0, half = 0.5d0, one=1.d0
    integer :: i97, j97, ivec, i, j, k, l, m, ij, kl, len, ii, jj
    logical :: test = .false.
    save test, i97, j97, c, cd, cm, u

    if (.not. test ) then
!     this is the initialization routine for the random number generator ranmar()
!     note: the seed variables can have values between:    0 <= ij <= 31328
!                                                          0 <= kl <= 30081
!     the random number sequences created by these two seeds are of sufficient
!     length to complete an entire calculation with. for example, if sveral
!     different groups are working on different parts of the same calculation,
!     each group could be assigned its own ij seed. this would leave each group
!     with 30000 choices for the second seed. that is to say, this random
!     number generator can create 900 million different subsequences -- with
!     each subsequence having a length of approximately 10^30.
!

      i = mod(ij/177, 177) + 2
      j = mod(ij    , 177) + 2
      k = mod(kl/169, 178) + 1
      l = mod(kl,     169)
      do ii = 1, 97
        s = zero
        t = half
        do jj = 1, 24
          m = mod(mod(i*j, 179)*k, 179)
          i = j
          j = k
          k = m
          l = mod(53*l+1, 169)
          if (mod(l*m, 64) .ge. 32) then
            s = s + t
            endif
          t = half * t
          end do
        u(ii) = s
        end do

      c = 362436.0 / 16777216.0
      cd = 7654321.0 / 16777216.0
      cm = 16777213.0 /16777216.0
      i97 = 97
      j97 = 33
      test = .true.
      endif

    do ivec = 1, len
      uni = u(i97) - u(j97)
      if( uni .lt. zero ) uni = uni + one
      u(i97) = uni
      i97 = i97 - 1
      if(i97 .eq. 0) i97 = 97
      j97 = j97 - 1
      if(j97 .eq. 0) j97 = 97
      c = c - cd
      if( c .lt. zero ) c = c + cm
      uni = uni - c
      if( uni .lt. zero ) uni = uni + one
      rvec(ivec) = uni
      end do
    return
    end subroutine ranmar

!============================================================================
    subroutine matvec (matr, vect, result, nord)
!
!   Utility to multiple a square matrix by a vector
!
    implicit none
    real(8) :: matr(nord,nord), vect(nord), result(nord), work(nord)
    integer :: nord, i, j

    do i = 1, nord
      work(i) = dot_product(matr(i,:), vect)
      end do
    result = work
    end subroutine matvec
